/*
 * Copyright 2018 ConsenSys AG.
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 */
package tech.pegasys.pantheon.consensus.ibft.statemachine;

import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.util.Lists.newArrayList;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.atLeastOnce;
import static org.mockito.Mockito.never;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.verifyNoMoreInteractions;
import static org.mockito.Mockito.when;

import tech.pegasys.pantheon.consensus.ibft.ConsensusRoundIdentifier;
import tech.pegasys.pantheon.consensus.ibft.IbftGossip;
import tech.pegasys.pantheon.consensus.ibft.ibftevent.BlockTimerExpiry;
import tech.pegasys.pantheon.consensus.ibft.ibftevent.IbftReceivedMessageEvent;
import tech.pegasys.pantheon.consensus.ibft.ibftevent.NewChainHead;
import tech.pegasys.pantheon.consensus.ibft.ibftevent.RoundExpiry;
import tech.pegasys.pantheon.consensus.ibft.messagedata.CommitMessageData;
import tech.pegasys.pantheon.consensus.ibft.messagedata.IbftV2;
import tech.pegasys.pantheon.consensus.ibft.messagedata.NewRoundMessageData;
import tech.pegasys.pantheon.consensus.ibft.messagedata.PrepareMessageData;
import tech.pegasys.pantheon.consensus.ibft.messagedata.ProposalMessageData;
import tech.pegasys.pantheon.consensus.ibft.messagedata.RoundChangeMessageData;
import tech.pegasys.pantheon.consensus.ibft.payload.CommitPayload;
import tech.pegasys.pantheon.consensus.ibft.payload.NewRoundPayload;
import tech.pegasys.pantheon.consensus.ibft.payload.PreparePayload;
import tech.pegasys.pantheon.consensus.ibft.payload.ProposalPayload;
import tech.pegasys.pantheon.consensus.ibft.payload.RoundChangePayload;
import tech.pegasys.pantheon.consensus.ibft.payload.SignedData;
import tech.pegasys.pantheon.ethereum.chain.Blockchain;
import tech.pegasys.pantheon.ethereum.core.Address;
import tech.pegasys.pantheon.ethereum.core.BlockHeader;
import tech.pegasys.pantheon.ethereum.p2p.api.Message;
import tech.pegasys.pantheon.ethereum.p2p.wire.DefaultMessage;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableMap;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.Mock;
import org.mockito.junit.MockitoJUnitRunner;

@RunWith(MockitoJUnitRunner.class)
public class IbftControllerTest {
  @Mock private Blockchain blockChain;
  @Mock private IbftFinalState ibftFinalState;
  @Mock private IbftBlockHeightManagerFactory blockHeightManagerFactory;
  @Mock private BlockHeader blockHeader;
  @Mock private IbftBlockHeightManager blockHeightManager;

  @Mock private SignedData<ProposalPayload> signedProposal;
  private Message proposalMessage;
  @Mock private ProposalMessageData proposalMessageData;
  @Mock private ProposalPayload proposalPayload;

  @Mock private SignedData<PreparePayload> signedPrepare;
  private Message prepareMessage;
  @Mock private PrepareMessageData prepareMessageData;
  @Mock private PreparePayload preparePayload;

  @Mock private SignedData<CommitPayload> signedCommit;
  private Message commitMessage;
  @Mock private CommitMessageData commitMessageData;
  @Mock private CommitPayload commitPayload;

  @Mock private SignedData<NewRoundPayload> signedNewRound;
  private Message newRoundMessage;
  @Mock private NewRoundMessageData newRoundMessageData;
  @Mock private NewRoundPayload newRoundPayload;

  @Mock private SignedData<RoundChangePayload> signedRoundChange;
  private Message roundChangeMessage;
  @Mock private RoundChangeMessageData roundChangeMessageData;
  @Mock private RoundChangePayload roundChangePayload;

  private final Map<Long, List<Message>> futureMessages = new HashMap<>();
  private final Address validator = Address.fromHexString("0x0");
  private final Address unknownValidator = Address.fromHexString("0x2");
  private final ConsensusRoundIdentifier futureRoundIdentifier = new ConsensusRoundIdentifier(2, 0);
  private ConsensusRoundIdentifier roundIdentifier = new ConsensusRoundIdentifier(0, 0);
  @Mock private IbftGossip ibftGossip;
  private IbftController ibftController;

  @Before
  public void setup() {
    when(blockChain.getChainHeadHeader()).thenReturn(blockHeader);
    when(blockHeightManagerFactory.create(blockHeader)).thenReturn(blockHeightManager);
    when(ibftFinalState.getValidators()).thenReturn(ImmutableList.of(validator));
    ibftController =
        new IbftController(
            blockChain, ibftFinalState, blockHeightManagerFactory, futureMessages, ibftGossip);
  }

  @Test
  public void createsNewBlockHeightManagerWhenStarted() {
    ibftController.start();
    assertThat(futureMessages).isEmpty();
    verify(blockHeightManagerFactory).create(blockHeader);
  }

  @Test
  public void startsNewBlockHeightManagerAndReplaysFutureMessages() {
    final ConsensusRoundIdentifier roundIdentifierHeight3 = new ConsensusRoundIdentifier(3, 0);
    setupPrepare(futureRoundIdentifier, validator);
    setupProposal(roundIdentifierHeight3, validator);
    setupCommit(futureRoundIdentifier, validator);
    setupRoundChange(futureRoundIdentifier, validator);
    setupNewRound(roundIdentifierHeight3, validator);

    final List<Message> height2Msgs =
        newArrayList(prepareMessage, commitMessage, roundChangeMessage);
    final List<Message> height3Msgs = newArrayList(proposalMessage, newRoundMessage);
    futureMessages.put(2L, height2Msgs);
    futureMessages.put(3L, height3Msgs);
    when(blockHeightManager.getChainHeight()).thenReturn(2L);

    ibftController.start();
    assertThat(futureMessages.keySet()).hasSize(1);
    assertThat(futureMessages.get(3L)).isEqualTo(height3Msgs);
    verify(blockHeightManagerFactory).create(blockHeader);
    verify(blockHeightManager, atLeastOnce()).getChainHeight();
    verify(blockHeightManager).start();
    verify(blockHeightManager, never()).handleProposalPayload(signedProposal);
    verify(blockHeightManager).handlePreparePayload(signedPrepare);
    verify(ibftGossip).gossipMessage(prepareMessage);
    verify(blockHeightManager).handleCommitPayload(signedCommit);
    verify(ibftGossip).gossipMessage(commitMessage);
    verify(blockHeightManager).handleRoundChangePayload(signedRoundChange);
    verify(ibftGossip).gossipMessage(roundChangeMessage);
    verify(blockHeightManager, never()).handleNewRoundPayload(signedNewRound);
  }

  @Test
  public void createsNewBlockHeightManagerAndReplaysFutureMessagesOnNewChainHeadEvent() {
    setupPrepare(futureRoundIdentifier, validator);
    setupProposal(futureRoundIdentifier, validator);
    setupCommit(futureRoundIdentifier, validator);
    setupRoundChange(futureRoundIdentifier, validator);
    setupNewRound(futureRoundIdentifier, validator);

    futureMessages.put(
        2L,
        ImmutableList.of(
            prepareMessage, proposalMessage, commitMessage, roundChangeMessage, newRoundMessage));
    when(blockHeightManager.getChainHeight()).thenReturn(2L);

    final NewChainHead newChainHead = new NewChainHead(blockHeader);
    ibftController.handleNewBlockEvent(newChainHead);

    verify(blockHeightManagerFactory).create(blockHeader);
    verify(blockHeightManager, atLeastOnce()).getChainHeight();
    verify(blockHeightManager).start();
    verify(blockHeightManager).handleProposalPayload(signedProposal);
    verify(ibftGossip).gossipMessage(proposalMessage);
    verify(blockHeightManager).handlePreparePayload(signedPrepare);
    verify(ibftGossip).gossipMessage(prepareMessage);
    verify(blockHeightManager).handleCommitPayload(signedCommit);
    verify(ibftGossip).gossipMessage(commitMessage);
    verify(blockHeightManager).handleRoundChangePayload(signedRoundChange);
    verify(ibftGossip).gossipMessage(roundChangeMessage);
    verify(blockHeightManager).handleNewRoundPayload(signedNewRound);
    verify(ibftGossip).gossipMessage(newRoundMessage);
  }

  @Test
  public void handlesRoundExpiry() {
    final RoundExpiry roundExpiry = new RoundExpiry(roundIdentifier);

    ibftController.start();
    ibftController.handleRoundExpiry(roundExpiry);

    verify(blockHeightManager).roundExpired(roundExpiry);
  }

  @Test
  public void handlesBlockTimerExpiry() {
    final BlockTimerExpiry blockTimerExpiry = new BlockTimerExpiry(roundIdentifier);

    ibftController.start();
    ibftController.handleBlockTimerExpiry(blockTimerExpiry);

    verify(blockHeightManager).handleBlockTimerExpiry(roundIdentifier);
  }

  @Test
  public void proposalForCurrentHeightIsPassedToBlockHeightManager() {
    setupProposal(roundIdentifier, validator);
    ibftController.start();
    ibftController.handleMessageEvent(new IbftReceivedMessageEvent(proposalMessage));

    assertThat(futureMessages).isEmpty();
    verify(blockHeightManager).handleProposalPayload(signedProposal);
    verify(ibftGossip).gossipMessage(proposalMessage);
    verify(blockHeightManager, atLeastOnce()).getChainHeight();
    verify(blockHeightManager).start();
    verifyNoMoreInteractions(blockHeightManager);
  }

  @Test
  public void prepareForCurrentHeightIsPassedToBlockHeightManager() {
    setupPrepare(roundIdentifier, validator);
    ibftController.start();
    ibftController.handleMessageEvent(new IbftReceivedMessageEvent(prepareMessage));

    assertThat(futureMessages).isEmpty();
    verify(blockHeightManager).handlePreparePayload(signedPrepare);
    verify(ibftGossip).gossipMessage(prepareMessage);
    verify(blockHeightManager, atLeastOnce()).getChainHeight();
    verify(blockHeightManager).start();
    verifyNoMoreInteractions(blockHeightManager);
  }

  @Test
  public void commitForCurrentHeightIsPassedToBlockHeightManager() {
    setupCommit(roundIdentifier, validator);
    ibftController.start();
    ibftController.handleMessageEvent(new IbftReceivedMessageEvent(commitMessage));

    assertThat(futureMessages).isEmpty();
    verify(blockHeightManager).handleCommitPayload(signedCommit);
    verify(ibftGossip).gossipMessage(commitMessage);
    verify(blockHeightManager, atLeastOnce()).getChainHeight();
    verify(blockHeightManager).start();
    verifyNoMoreInteractions(blockHeightManager);
  }

  @Test
  public void newRoundForCurrentHeightIsPassedToBlockHeightManager() {
    roundIdentifier = new ConsensusRoundIdentifier(0, 1);
    setupNewRound(roundIdentifier, validator);
    ibftController.start();
    ibftController.handleMessageEvent(new IbftReceivedMessageEvent(newRoundMessage));

    assertThat(futureMessages).isEmpty();
    verify(blockHeightManager).handleNewRoundPayload(signedNewRound);
    verify(ibftGossip).gossipMessage(newRoundMessage);
    verify(blockHeightManager, atLeastOnce()).getChainHeight();
    verify(blockHeightManager).start();
    verifyNoMoreInteractions(blockHeightManager);
  }

  @Test
  public void roundChangeForCurrentHeightIsPassedToBlockHeightManager() {
    roundIdentifier = new ConsensusRoundIdentifier(0, 1);
    setupRoundChange(roundIdentifier, validator);
    ibftController.start();
    ibftController.handleMessageEvent(new IbftReceivedMessageEvent(roundChangeMessage));

    assertThat(futureMessages).isEmpty();
    verify(blockHeightManager).handleRoundChangePayload(signedRoundChange);
    verify(ibftGossip).gossipMessage(roundChangeMessage);
    verify(blockHeightManager, atLeastOnce()).getChainHeight();
    verify(blockHeightManager).start();
    verifyNoMoreInteractions(blockHeightManager);
  }

  @Test
  public void proposalForPastHeightIsDiscarded() {
    setupProposal(roundIdentifier, validator);
    when(blockHeightManager.getChainHeight()).thenReturn(1L);
    verifyNotHandledAndNoFutureMsgs(new IbftReceivedMessageEvent(proposalMessage));
  }

  @Test
  public void prepareForPastHeightIsDiscarded() {
    setupPrepare(roundIdentifier, validator);
    when(blockHeightManager.getChainHeight()).thenReturn(1L);
    verifyNotHandledAndNoFutureMsgs(new IbftReceivedMessageEvent(prepareMessage));
  }

  @Test
  public void commitForPastHeightIsDiscarded() {
    setupCommit(roundIdentifier, validator);
    when(blockHeightManager.getChainHeight()).thenReturn(1L);
    verifyNotHandledAndNoFutureMsgs(new IbftReceivedMessageEvent(commitMessage));
  }

  @Test
  public void newRoundForPastHeightIsDiscarded() {
    setupNewRound(roundIdentifier, validator);
    when(blockHeightManager.getChainHeight()).thenReturn(1L);
    verifyNotHandledAndNoFutureMsgs(new IbftReceivedMessageEvent(newRoundMessage));
  }

  @Test
  public void roundChangeForPastHeightIsDiscarded() {
    setupRoundChange(roundIdentifier, validator);
    when(blockHeightManager.getChainHeight()).thenReturn(1L);
    verifyNotHandledAndNoFutureMsgs(new IbftReceivedMessageEvent(roundChangeMessage));
  }

  @Test
  public void roundExpiryForPastHeightIsDiscarded() {
    final RoundExpiry roundExpiry = new RoundExpiry(roundIdentifier);
    when(blockHeightManager.getChainHeight()).thenReturn(1L);
    ibftController.start();
    ibftController.handleRoundExpiry(roundExpiry);
    assertThat(futureMessages).isEmpty();
    verify(blockHeightManager, never()).roundExpired(any());
  }

  @Test
  public void blockTimerForPastHeightIsDiscarded() {
    final BlockTimerExpiry blockTimerExpiry = new BlockTimerExpiry(roundIdentifier);
    when(blockHeightManager.getChainHeight()).thenReturn(1L);
    ibftController.start();
    ibftController.handleBlockTimerExpiry(blockTimerExpiry);
    assertThat(futureMessages).isEmpty();
    verify(blockHeightManager, never()).handleBlockTimerExpiry(any());
  }

  @Test
  public void proposalForUnknownValidatorIsDiscarded() {
    setupProposal(roundIdentifier, unknownValidator);
    verifyNotHandledAndNoFutureMsgs(new IbftReceivedMessageEvent(proposalMessage));
  }

  @Test
  public void prepareForUnknownValidatorIsDiscarded() {
    setupPrepare(roundIdentifier, unknownValidator);
    verifyNotHandledAndNoFutureMsgs(new IbftReceivedMessageEvent(prepareMessage));
  }

  @Test
  public void commitForUnknownValidatorIsDiscarded() {
    setupCommit(roundIdentifier, unknownValidator);
    verifyNotHandledAndNoFutureMsgs(new IbftReceivedMessageEvent(commitMessage));
  }

  @Test
  public void newRoundForUnknownValidatorIsDiscarded() {
    setupNewRound(roundIdentifier, unknownValidator);
    verifyNotHandledAndNoFutureMsgs(new IbftReceivedMessageEvent(newRoundMessage));
  }

  @Test
  public void roundChangeForUnknownValidatorIsDiscarded() {
    setupRoundChange(roundIdentifier, unknownValidator);
    verifyNotHandledAndNoFutureMsgs(new IbftReceivedMessageEvent(roundChangeMessage));
  }

  @Test
  public void proposalForFutureHeightIsBuffered() {
    setupProposal(futureRoundIdentifier, validator);
    final Map<Long, List<Message>> expectedFutureMsgs =
        ImmutableMap.of(2L, ImmutableList.of(proposalMessage));
    verifyHasFutureMessages(new IbftReceivedMessageEvent(proposalMessage), expectedFutureMsgs);
  }

  @Test
  public void prepareForFutureHeightIsBuffered() {
    setupPrepare(futureRoundIdentifier, validator);
    final Map<Long, List<Message>> expectedFutureMsgs =
        ImmutableMap.of(2L, ImmutableList.of(prepareMessage));
    verifyHasFutureMessages(new IbftReceivedMessageEvent(prepareMessage), expectedFutureMsgs);
  }

  @Test
  public void commitForFutureHeightIsBuffered() {
    setupCommit(futureRoundIdentifier, validator);
    final Map<Long, List<Message>> expectedFutureMsgs =
        ImmutableMap.of(2L, ImmutableList.of(commitMessage));
    verifyHasFutureMessages(new IbftReceivedMessageEvent(commitMessage), expectedFutureMsgs);
  }

  @Test
  public void newRoundForFutureHeightIsBuffered() {
    setupNewRound(futureRoundIdentifier, validator);
    final Map<Long, List<Message>> expectedFutureMsgs =
        ImmutableMap.of(2L, ImmutableList.of(newRoundMessage));
    verifyHasFutureMessages(new IbftReceivedMessageEvent(newRoundMessage), expectedFutureMsgs);
  }

  @Test
  public void roundChangeForFutureHeightIsBuffered() {
    setupRoundChange(futureRoundIdentifier, validator);
    final Map<Long, List<Message>> expectedFutureMsgs =
        ImmutableMap.of(2L, ImmutableList.of(roundChangeMessage));
    verifyHasFutureMessages(new IbftReceivedMessageEvent(roundChangeMessage), expectedFutureMsgs);
  }

  private void verifyNotHandledAndNoFutureMsgs(final IbftReceivedMessageEvent msg) {
    ibftController.start();
    ibftController.handleMessageEvent(msg);

    assertThat(futureMessages).isEmpty();
    verify(blockHeightManager, atLeastOnce()).getChainHeight();
    verify(blockHeightManager).start();
    verifyNoMoreInteractions(blockHeightManager);
  }

  private void verifyHasFutureMessages(
      final IbftReceivedMessageEvent msg, final Map<Long, List<Message>> expectedFutureMsgs) {
    ibftController.start();
    ibftController.handleMessageEvent(msg);

    assertThat(futureMessages).hasSize(expectedFutureMsgs.size());
    assertThat(futureMessages).isEqualTo(expectedFutureMsgs);
    verify(blockHeightManager, atLeastOnce()).getChainHeight();
    verify(blockHeightManager).start();
    verifyNoMoreInteractions(blockHeightManager);
  }

  private void setupProposal(
      final ConsensusRoundIdentifier roundIdentifier, final Address validator) {
    when(signedProposal.getPayload()).thenReturn(proposalPayload);
    when(signedProposal.getSender()).thenReturn(validator);
    when(proposalPayload.getRoundIdentifier()).thenReturn(roundIdentifier);
    when(proposalMessageData.getCode()).thenReturn(IbftV2.PROPOSAL);
    when(proposalMessageData.decode()).thenReturn(signedProposal);
    proposalMessage = new DefaultMessage(null, proposalMessageData);
  }

  private void setupPrepare(
      final ConsensusRoundIdentifier roundIdentifier, final Address validator) {
    when(signedPrepare.getPayload()).thenReturn(preparePayload);
    when(signedPrepare.getSender()).thenReturn(validator);
    when(preparePayload.getRoundIdentifier()).thenReturn(roundIdentifier);
    when(prepareMessageData.getCode()).thenReturn(IbftV2.PREPARE);
    when(prepareMessageData.decode()).thenReturn(signedPrepare);
    prepareMessage = new DefaultMessage(null, prepareMessageData);
  }

  private void setupCommit(
      final ConsensusRoundIdentifier roundIdentifier, final Address validator) {
    when(signedCommit.getPayload()).thenReturn(commitPayload);
    when(signedCommit.getSender()).thenReturn(validator);
    when(commitPayload.getRoundIdentifier()).thenReturn(roundIdentifier);
    when(commitMessageData.getCode()).thenReturn(IbftV2.COMMIT);
    when(commitMessageData.decode()).thenReturn(signedCommit);
    commitMessage = new DefaultMessage(null, commitMessageData);
  }

  private void setupNewRound(
      final ConsensusRoundIdentifier roundIdentifier, final Address validator) {
    when(signedNewRound.getPayload()).thenReturn(newRoundPayload);
    when(signedNewRound.getSender()).thenReturn(validator);
    when(newRoundPayload.getRoundIdentifier()).thenReturn(roundIdentifier);
    when(newRoundMessageData.getCode()).thenReturn(IbftV2.NEW_ROUND);
    when(newRoundMessageData.decode()).thenReturn(signedNewRound);
    newRoundMessage = new DefaultMessage(null, newRoundMessageData);
  }

  private void setupRoundChange(
      final ConsensusRoundIdentifier roundIdentifier, final Address validator) {
    when(signedRoundChange.getPayload()).thenReturn(roundChangePayload);
    when(signedRoundChange.getSender()).thenReturn(validator);
    when(roundChangePayload.getRoundIdentifier()).thenReturn(roundIdentifier);
    when(roundChangeMessageData.getCode()).thenReturn(IbftV2.ROUND_CHANGE);
    when(roundChangeMessageData.decode()).thenReturn(signedRoundChange);
    roundChangeMessage = new DefaultMessage(null, roundChangeMessageData);
  }
}
